# When running on Windows in RStudio, we need to patch subprocess.Popen
# https://github.com/rstudio/reticulate/issues/1448


def patch_subprocess_Popen():
    from functools import wraps
    import subprocess

    og_Popen__init__ = subprocess.Popen.__init__

    @wraps(subprocess.Popen.__init__)
    def __init__(self, *args, **kwargs):
        if kwargs.get("stdin") is None:
            kwargs["stdin"] = subprocess.DEVNULL
        return og_Popen__init__(self, *args, **kwargs)

    subprocess.Popen.__init__ = __init__



import rpycall
import threading

import sys

is_py2 = sys.version[0] == "2"
if is_py2:
    import Queue as queue
else:
    import queue as queue

_completed_sentinel = object()


class RGenerator(object):
    def __init__(self, r_function, prefetch=0):
        self.r_function = r_function
        self.prefetch = prefetch
        self.values_queue = queue.Queue()
        self.completed = False
        self._pending_tend_queue = False
        if prefetch:
            self._tend_queue()

    def __iter__(self):
        return self

    def next(self):
        return self.__next__()

    def __next__(self):
        self._tend_queue(1)
        val = self._get_or_raise()
        if self.prefetch:
            self._tend_queue()
        return val

    def _tend_queue(self, min_fetch=0):
        if self.completed:
            return

        ## Prefetch and enqueue generator generated values.

        # If we're not on the main thread, make sure there is a pending call to
        # this function from the main thread and return.
        if threading.current_thread() is not threading.main_thread():
            if not self._pending_tend_queue:
                self._pending_tend_queue = True
                rpycall.call_python_function_on_main_thread(self._tend_queue, min_fetch)
            return

        # We're on the main thread, call the generator and put values on the queue.
        self._pending_tend_queue = False

        fetch = max(min_fetch, self.prefetch - self.values_queue.qsize())
        for _ in range(fetch):
            try:
                val = self.r_function()
            except StopIteration:
                self.values_queue.put(_completed_sentinel)
                self.completed = True
                return

            self.values_queue.put(val)

    def _get_or_raise(self):
        try:
            # only wait/block-thread/yield-to-another-thread if the R generator
            # is not exhausted.
            val = self.values_queue.get(block=~self.completed)
            if val is _completed_sentinel:
                raise StopIteration()
            return val
        except queue.Empty:
            # only get here if self.completed = True and the queue is empty,
            # meaning we've already pulled off _completed_sentinel and
            # raised StopIteration
            raise StopIteration()



import rpycall


def make_python_function(f, name=None):
    def python_function(*args, **kwargs):
        # call the function
        value, error = rpycall.call_r_function(f, *args, **kwargs)

        if error:
            if isinstance(error, str) and error == "KeyboardInterrupt":
                # Only reachable if a C++ exception was caught in call_r_function()
                # otherwise error is always an Exception object
                raise KeyboardInterrupt()

            if isinstance(error, BaseException):
                raise error

            # basically unreachable since R errors get automatically
            # converted to python Exceptions, but just in case
            raise RuntimeError(error)
        return value

    if not name is None:
        python_function.__name__ = name

    return python_function




import sys
import threading

if sys.version_info.major < 3:
  import __builtin__ as builtins
else:
  import builtins


# The R callback to be run. Initialized in the 'initialize()' method.
_callback = None

# A list of Python packages which have been imported. The aforementioned
# callback will be run on the main thread after a module has been imported
# on the main thread.
_imported_packages = []

# A simple counter, tracking the recursion depth. This is used as we only
# attempt to run the R callback at the top level; that is, we don't want
# to run it while modules are being loaded recursively.
_recursion_depth = 0

# The builtin implementation of '__import__'; saved so that we can re-use it
# after initialization.
__import__ = builtins.__import__

# The implementation of '_find_and_load' captured from 'importlib._bootstrap'.
# Since we're trying to poke at Python internals, we try to wrap this code
# in try-catch and only use this if it appears safe to do so.
_find_and_load = None
try:
  import importlib._bootstrap
  _find_and_load = importlib._bootstrap._find_and_load
except:
  pass


# Run hooks on imported packages, if safe to do so.
def _maybe_run_hooks():

  # Don't run hooks while loading packages recursively.
  global _recursion_depth
  if _recursion_depth != 0:
    return False

  # Check whether we're on the main thread. Note that separate threads can
  # attempt to load Python modules, but the R callback we register can only
  # be safely run on the main thread.
  is_main_thread = isinstance(threading.current_thread(), threading._MainThread)
  if not is_main_thread:
    return False

  # Pre-flight checks passed; run the callbacks.
  global _imported_packages
  global _callback
  for package in _imported_packages:
    _callback(package)

  # Clear the import list.
  del _imported_packages[:]


# Resolve a module name on import. See Python code here for motivation.
# https://github.com/python/cpython/blob/c5140945c723ae6c4b7ee81ff720ac8ea4b52cfd/Lib/importlib/_bootstrap.py#L1246-L1270
def _resolve_module_name(name, globals=None, level=0):

  if level == 0:
    return name

  package = globals.get("__package__")
  if package is not None:
    return package

  spec = globals.get("__spec__")
  if spec is not None:
    return spec.parent
  
  return name

# Helper function for running an import hook with our extra scaffolding.
def _run_hook(name, hook):
  
  # Check whether this module has already been imported.
  already_imported = name in sys.modules

  # Bump the recursion depth.
  global _recursion_depth
  _recursion_depth += 1

  # Run the hook.
  try:
    module = hook()
  except:
    raise
  finally:
    _recursion_depth -= 1

  # Add this package to the import list, if this is the first
  # time importing that package.
  global _imported_packages
  if not already_imported:
    _imported_packages.append(name)

  # try and run hooks if possible
  _maybe_run_hooks()

  # return loaded module
  return module


# The hook installed to replace 'importlib._bootstrap._find_and_load'.
def _find_and_load_hook(name, import_):

  def _hook():
    global _find_and_load
    return _find_and_load(name, import_)
  
  return _run_hook(name, _hook)
  
# Initialize the '_find_and_load' replacement hook.
def _initialize_importlib():

  import importlib._bootstrap
  importlib._bootstrap._find_and_load = _find_and_load_hook
  

# The hook installed to replace '__import__'.
def _import_hook(name, globals=None, locals=None, fromlist=(), level=0):

  # resolve module name
  resolved_module_name = _resolve_module_name(name, globals, level)
  
  def _hook():
    global __import__
    return __import__(name, globals=globals, locals=locals, fromlist=fromlist, level=level)
  
  return _run_hook(_hook)

# Initialize the '__import__' hook.
def _initialize_default():
  builtins.__import__ = _import_hook


# The main entrypoint for this module.
def initialize(callback):

  # Save the callback.
  global _callback
  _callback = callback

  # Check whether we can initialie with importlib.
  global _find_and_load
  if _find_and_load is not None:
    return _initialize_importlib()

  # Otherwise, fall back to default implementation.  
  return _initialize_default()
